#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#if defined(Linux)                                                              
#include <unistd.h>                                                             
#endif                                                                          

#include <rlcif.h>
#include "rmdpif.h"

#define USAGE \
	"Usage:\n\t RMDPft -[tifCGxgep] m.cast.ip/port/ttl"		\
	"\n\t RMDPft -s -[tifCGxgI] m.cast.ip/port/ttl\n"		\
	"options:\n"							\
	"\t-s \t\t\t Is sender (receiver by default)\n"			\
	"\t-i <session id>\n"						\
	"\t-f <I/O filename>\n"						\
	"\t-C <session name>\n"						\
	"\t-G \t\t\t Use GUI (only if compiled in)\n"			\
	"\t-x \t\t\t Ask for filename\n"				\
	"\t-g <window geometry>\n"					\
	"\t-e <event-log file> \t (receiver only)\n"			\
	"\t-p <poll-log file> \t (receiver only)\n"			\
	"\t-I \t\t\t Sender never stops\n"				\
	"\t-m \t\t\t Multicast feedback, def.is Unicast (rec.only)\n"	\
	"\t-F <stats file> \t Log file for stats report (sender only)\n"\
	"\t-h \t\t\t Prints the usage message\n"

static int isreceiver = 1;	/* receiver by default */
#define ISR	((isreceiver>0)?1:0)
int rlc_i;			/* rlc instance id */
int rmdp_i;			/* rmdp instance id */
int fd = -1;			/* IO file */
int chosefile = 0;
char *sessname=NULL;
char *geometry=NULL;
#ifdef HAVE_UI
int guicompiled = 1;
#else
int guicompiled = 0;
#endif


void
badcmdline(char *arg)
{
    rmdp_my_error(-1,
	"Bad argument %s\n"
	USAGE
	,
	arg);
	/* "\t-n <number of layers>\n" */
	/* "\t-d <M-cast data address/port>\n" */
}

int open_n_check(char *filn, int wr)
{
  int fd = -1;
#ifdef HAVE_UI
  int res;
  char *tit, *ret = NULL;
#endif

    if  (!strcmp(filn, "-")) {
	if (wr)
	    return 1;
	else
	    return 0;
    } else if  (!strcmp(filn, "-GUI") && !wr) {
#ifdef HAVE_UI
	if (sessname)
	    tit = sessname;
	else
	    tit = "RMDP";
	res = rlc_gui_tcl_exec (rlc_i, &ret,
	    	"tk_getOpenFile -title {%s : Open File}", tit);
	if (res >= 0 && ret != NULL && ret[0] != '\0')
	    filn = ret;
#else
	return -1;
#endif
    }

    if  ((wr && (fd = open (filn, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0) ||
	 (!wr && (fd = open (filn, O_RDONLY)) < 0)) {
	return (-1);
    } else {
	return (fd);
    }
}

void install_help ()
{
#ifdef HAVE_UI
  char *ret;
    if ( rlc_gui_tcl_exec (rlc_i, &ret,
	"set RMDPftINFOS {" USAGE "}") == 0) {
	if (rlc_gui_tcl_exec (rlc_i, &ret,
	    "if [info exist HelpButt] {"
	    "$HelpButt add command -label \"About RMDPft\" -command "
	    "\"show_infos {$RMDPftINFOS} \\\"About RMDPft\\\"\""
	    "}") != 0)
	    fprintf(stderr, "** add command returns : %s\n", ret);
    } else
	fprintf(stderr, "** set RMDPftINFOS returns : %s\n", ret);
#endif
}

/* this handle the termination of
 * the current reception */
void reception_end(int rmdp_i, int success)
{
    if (success >= 0 && chosefile) {
#ifdef HAVE_UI
	char *ret = NULL;
	int res;
	if (sessname)
	    res = rlc_gui_tcl_exec (rlc_i, &ret,
		"tk_getSaveFile -title {%s : Save File}", sessname);
	else
	    res = rlc_gui_tcl_exec (rlc_i, &ret,
	    	"tk_getSaveFile -title \"RMDP : Save File\"");
	if (res >= 0 && ret != NULL) {
	    if (ret[0] != '\0') {
		if ((fd = open_n_check(ret, ISR)) >= 0) {
		    rmdp_set_io(rmdp_i, fd);
		} else
		    rmdp_my_error (rmdp_i,
			"Error opening out file (%s)\n", ret);
	    } else {
		rmdp_my_error (rmdp_i, "No file selected!\n");
	    }
	    free(ret);
	} else {
	    rmdp_my_error (rmdp_i, "rlc gui error|\n");
	}
#endif
    } else if (success < 0) {
	rmdp_my_error (rmdp_i, "Decoding error ? \n");
    }
    /* in this example just abort the reception instance */
    rlc_abort(rlc_i);
    /* rmdp_delete(rmdp_i); */
}

int
main(int argc, char *argv[])
{

  char *dataaddr = NULL;
  char *tmpfilename = NULL;
  char *statsrepfile = NULL;
  int tttl, ttl = 5;
  int have_gui = 0;
  int multifeed = 0;
  int stafd = -1;

  int ret;

  char *ev_f = NULL , *po_f = NULL ;
  char tmpaddr[512];
  unsigned long addr0;
  short port;
  int pkt_size;		/* rmdp payload + rmdp header */
  int pid = -1;

  addr_list_t taddr;
  rmdp_addr_t adds;

  int errflg = 0, i;

  rmdp_rx_packet_t rx_f;

  int layers;				/* # of layers */
  unsigned long base_rate = 16 * 1024;	/* input parms ? */
  unsigned long p;
  int nostop = 0;
  rlc_tx_params_t parml;



  extern char *optarg;
  extern int optind;

    /* parse command line... */
    while ((i = getopt(argc,argv,"hsd:t:i:f:C:Gxg:Ime:p:F:")) != EOF)
	switch(i) {

	case 'h': /* help */
	    badcmdline("?");
	    break;
	case 's': /* is sender ! */
	    isreceiver = 0;
	    break;
	case 'd': /* data addr/port */
	    if ((dataaddr = malloc(strlen(optarg)+1)) == NULL)
		break;
	    strncpy(dataaddr, optarg, strlen(optarg)+1);
	    break;
	case 't': /* TTL */
	    ttl = atoi(optarg);
	    if (ttl <= 0 || ttl > 255) badcmdline(optarg);
	    break;
	case 'i': /* pid */
	    pid = atoi(optarg);
	    break;
	case 'f': /* input/output file */
	    if ((tmpfilename = malloc(strlen(optarg)+1)) == NULL)
		break;
	    strncpy(tmpfilename, optarg, strlen(optarg)+1);
	    break;
	case 'C': /* session name */
	    if ((sessname = malloc(strlen(optarg)+1)) == NULL)
		break;
	    strcpy (sessname, optarg);
	    break;
	case 'G': /* have GUI */
	    have_gui = 1;
	    break;
	case 'x': /* GUI select I/O file */
	    chosefile = 1;
	    break;
	case 'g': /* geometry */
	    if ((geometry = malloc(strlen(optarg)+1)) == NULL)
		break;
	    strcpy (geometry, optarg);
	    break;
	case 'I': /* sender no-stop */
	    nostop = 1;
	    break;
	case 'm': /* multicast feedback */
	    multifeed = 1;
	    break;
	case 'e': /* event log file */
	    ev_f = (char *)calloc(strlen(optarg)+1, sizeof(char));
	    strncpy(ev_f, optarg, strlen(optarg)+1);
	    break;
	case 'p': /* poll log file */
	    po_f = (char *)calloc(strlen(optarg)+1, sizeof(char));
	    strncpy(po_f, optarg, strlen(optarg)+1);
	    break;
	case 'F': /* stats report file */
	    if ((statsrepfile = malloc(strlen(optarg)+1)) == NULL)
		break;
	    strncpy(statsrepfile, optarg, strlen(optarg)+1);
	    break;
	case '?':
	default:
	    errflg++;
	}

    if (errflg)
	badcmdline("?");

    argc -= optind;
    argv += optind;

    if (argc > 0 && (dataaddr = malloc(strlen(argv[0])+1)) != NULL) {
	strncpy(dataaddr, argv[0], strlen(argv[0])+1);
    }

    /* and now some checks... */
    if (have_gui && !guicompiled)
	badcmdline("GUI is not compiled in!");
    if (!dataaddr)
	badcmdline("must provide destination address");
    if ( !chosefile &&
	((!tmpfilename && (fd = open_n_check("-", ISR)) < 0) ||
	 (tmpfilename && (fd = open_n_check(tmpfilename, ISR)) < 0)) ) {
	badcmdline("input file name");
    }
    if (isreceiver && nostop)
	badcmdline("no-stop option only for the sender");
    if (!isreceiver && (ev_f || po_f))
	badcmdline("logging facility only in the receiver");
    if (!have_gui && (chosefile || geometry))
	badcmdline("GUI is disabled!");
    if (isreceiver && statsrepfile)
	badcmdline("-F option only for the sender");
    if (statsrepfile && (stafd = open_n_check(statsrepfile, 1)) < 0)
	badcmdline("cannot open stats report file");


    /* extract address (string) and port (short) from the
     * input parameter (x.x.x.x/yyy) */
#if 1
    if (rmdp_n_parse_addr(dataaddr, tmpaddr, sizeof(tmpaddr), &port,
	&tttl, "data") < 0)
    {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1, "Error %d in rmdp_parse_addr (%s)\n", ret, c);
    }
    if (tttl >= 0 && tttl <= 255)
	ttl = tttl;
#else
    if (rmdp_parse_addr(dataaddr, tmpaddr, sizeof(tmpaddr), &port,
	"data") < 0)
    {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1, "Error %d in rmdp_parse_addr (%s)\n", ret, c);
    }
#endif
    addr0 = ntohl(inet_addr(tmpaddr));
    adds.addr = addr0;
    adds.port = port+1;
    adds.ttl = ttl;

    /* setup an rmdp receiver instance */
    if ((ret = rmdp_i = rmdp_new (isreceiver)) < 0 ||
	(ret = rmdp_set_addr (rmdp_i, &adds)) < 0 ||
	(fd >= 0 && (ret = rmdp_set_io (rmdp_i, fd)) < 0) ) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1,
	    "Error %d in protocol initialization (%s)\n", ret, c);
    }
    if (pid != -1 && (ret = rmdp_set_id(rmdp_i, pid)) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1,
	    "Error %d in protocol initialization (%s)\n", ret, c);
    }

    /* set reception-end callback */
    if (isreceiver) {
	if ((ret = rmdp_set_end_callback(rmdp_i, reception_end)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_my_error (-1, 
		"Error %d in rmdp_set_end_callback (%s)\n", ret, c);
	}
	if (multifeed && (ret = rmdp_rx_enable_multifeed(rmdp_i, 1)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_my_error (-1, 
		"Error %d in rmdp_rx_enable_multifeed (%s)\n", ret, c);
	}
    }

    /* get data reception/transmission handler function from rmdp */
    if ((ret = rmdp_get_f(rmdp_i, (void **)(&rx_f))) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1, 
	    "Error %d in rmdp_get_f (%s)\n", ret, c);
    }

    taddr.nadd = 1;
    taddr.addrs = &addr0;
    taddr.port = port;
    taddr.ttl = ttl;

    /* setup a rlc instance */
    if ((ret = rlc_i = rlc_new (isreceiver)) < 0 ||
	(ret = rlc_set_addr (rlc_i, &taddr)) < 0 ||
	(ret = rlc_attach (rlc_i, rmdp_i, (void *)rx_f)) < 0 ) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_delete(rmdp_i);
	rmdp_my_error(-1, "error %d in rlc_setup (%s)\n", ret, c);
    }

    if ((ret = rmdp_set_rlc (rmdp_i, rlc_i)) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1,
	    "Error %d in protocol initialization (%s)\n", ret, c);
    }

    if (!isreceiver) {
	/* retrieve rmdp packet size (rlc payload) */
	pkt_size = rmdp_get_psize(rmdp_i);
	/* retrieve rmdp number of layers */
	layers = rmdp_get_nlay(rmdp_i);
	/* this is the interpacket-period (us) in the upper session */
	p = 1e6 * (pkt_size) * 8 / base_rate / (1<<(layers-1));

	parml.nlay = layers;
	parml.mpkts = pkt_size;
	parml.Mpkts = pkt_size;
	parml.t = p;
	if ((ret = rlc_tx_setup(rlc_i, &parml)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_delete(rmdp_i);
	    rmdp_my_error(-1, "error %d in rlc_tx_setup (%s)\n", ret, c);
	}

	if (nostop) {
	    unsigned long left = rmdp_get_left (rmdp_i);
	    if ((ret = rmdp_incr_left (rmdp_i,
		RMDP_LEFT_INFINITE - left)) < 0) {
		char *c;
		ret = rlc_error_info(&c);
		rmdp_my_error (-1,
		    "Error %d in protocol initialization (%s)\n", ret, c);
	    }
	}
    }

    /* setup rlc idle watchdog */
    if (isreceiver && (ret = rlc_rx_set_watchdog(rlc_i, NULL, 0)) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_delete(rmdp_i);
	rmdp_my_error(-1, "error %d in rlc_rx_setup (%s)\n", ret, c);
    }

#ifdef HAVE_UI
    if (have_gui) {
	if (geometry)
	    rlc_gui_geometry (rlc_i, geometry);
	if ((ret = rlc_gui_enable(rlc_i)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_delete(rmdp_i);
	    rmdp_my_error(-1, "error %d in rlc_gui_enable (%s)\n", ret, c);
	}

	/* enable UI in rmdp */
	if ((ret = rmdp_enable_gui(rmdp_i)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_my_error (-1,
		"Error %d in protocol initialization (%s)\n", ret, c);
	}

	if (sessname != NULL &&
	    (ret = rmdp_set_name (rmdp_i, sessname)) < 0) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_my_error (-1,
		"Error %d in protocol initialization (%s)\n", ret, c);
	}

	install_help ();

	if (!isreceiver && fd < 0 && chosefile &&
	    ( (fd = open_n_check("-GUI", ISR)) < 0 ||
	      (ret = rmdp_set_io (rmdp_i, fd)) < 0 )
	   ) {
	    char *c;
	    ret = rlc_error_info(&c);
	    rmdp_my_error (-1,
		"Error %d in protocol initialization (%s)\n", ret, c);
	}
    }
#endif

    /* possibly setup rlc log facility */
    if (isreceiver && (ev_f || po_f))
	(void)rlc_rx_log(rlc_i, ev_f, po_f);

    if (isreceiver && (ret = rmdp_set_stat_reports (rmdp_i, 1) ) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1,
	    "Error %d in rmdp_set_stat_reports (%s)\n", ret, c);
    }

    if (!isreceiver && stafd >= 0 &&
	(ret = rmdp_set_stat_reports (rmdp_i, stafd)) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_my_error (-1,
	    "Error %d in rmdp_set_stat_reports (%s)\n", ret, c);
    }

    /* start the reception */
    if ((ret = rlc_start(rlc_i)) < 0) {
	char *c;
	ret = rlc_error_info(&c);
	rmdp_delete(rmdp_i);
	rmdp_my_error(-1,
	    "error %d in rlc_start (%s)\n", ret, c);
    }

    while (rlc_getevent(1)!=EV_TERMINATE);

    exit(0);
}
