#ifndef RMDP_AGENT_H
#define RMDP_AGENT_H

#include "rlcif.h"
#include "rmdpif.h"

extern void *rmdp_inst_mngr;

#define AGENT_FIELDS \
    INSTANCE_DES_MEMB							       \
    unsigned long sessid;	/* random session identifier */		       \
    void ***alldata;		/* packets... */			       \
    void ***allpkts;		/* all plain data packets */		       \
    int B, k;			/* encoding parameters */		       \
    unsigned long npkts, flength;					       \
    int psize;			/* pkt size */				       \
									       \
    void *feccode;							       \
    int rlc_pi;			/* rlc associated protocol instance */	       \
    int have_gui;		/* have graphyc interface ? */		       \
									       \
    unsigned long addr;		/* feedback channel address, net byte ord. */  \
    short port;			/* feedback channel port #, net byte ord. */   \
    int ufeeso;			/* unicast feedback socket */		       \
    int mfeeso;			/* multicast feedback socket */		       \
									       \
    int fd;			/* input/output file descriptor */	       \
    int isreceiver;		/* 1 if this is receiver, 0 sender */	       \
    rmdp_end_f end_f;		/* called when when the reception	       \
				   terminates successfully */

typedef struct _rmdp_inst_t {
    AGENT_FIELDS

    /* sender specific ... */
    int nlay;
    unsigned short *idxs;	/* counts the packet transmitted
				   in each layer */
    char *sparepkt;		/* spare pkt pointer */
    unsigned long pkts_left;	/* packets left to transmit */
    unsigned long init_left;	/* initial scheduled packets */
    int stafd;			/* fil descriptor for stats report */

    /* receiver specific ... */
    int ttl;
    int **whichdata, *countdata;/* info on received packets */

    int maxn;			/* max packet index among those received */
    int missingblocks, duppackets;
    int oldtotgot, totgot, netgot;
    double speedf;		/* statistic: # pkts got for each
				   pkt sent on lay 0 */


    rmdp_rx_err_f err_f;	/* called when when the reception
				   terminates beacause of an error */
    int reqpending;		/* if a transmission extension timeout is
				   pending, its > 0 */
    int reqsent;
    int recpercent;
    unsigned long left;
    int mcastfeed;		/* is != 0 uses mucast feedback otherwise
				   uses unicast */
} rmdp_inst_t;

void _rmdp_rx_packet(int ides, int lay, char flags, char *pkt,
        int size);
int _rmdp_tx_packet(int des, int lay, char flags, char **p);
void _rmdp_deleteinstance(rmdp_inst_t *inst);
int _rmdp_allocation(rmdp_inst_t *se);
void _rmdp_deallocation(rmdp_inst_t *inst);


#if 1
int _rmdp_insert_instance (rmdp_inst_t * inst);
void _rmdp_return_instance (rmdp_inst_t * inst);
rmdp_inst_t * _rmdp_getinstance(int pi);
void _rmdp_init_instance ();
#endif

#endif  /* RMDP_AGENT_H */
