#include "layers.h"
#include "rmdp.h"
#include "compat.h"
#include "assert.h"

#include <stdlib.h>
#include <stdio.h>

/*
 * Establish packet organization within layers.
 * `lay' is the layer index (0..nc-1) and `lidx' the
 * slot index within the layer. `idx' is the global
 * slot index.
 *
 * Compute pkt# and blk#. If `idx' (global index) == -1
 * use `lay' and `lidx' (layer, and layer index). Otherwise
 * use `idx' and compute `lay' as well, but not `lidx' (XXX).
 * In this case if lay >= 0 chack that its value matches with
 * the one computed.
 * Always return the global index.
 */
int
which_packet(int B, int k, int nc, int idx, int *lay,
#if 0	/*ppp*/
	 int *lidx, int *pkt, int *blk)
#else
	 unsigned short *lidx, int *pkt, int *blk)
#endif
{
    int x, R = 1<<(nc-1);

    if (idx < 0)	/* compute `idx' from `lay' and `lidx' */
	if (*lay == 0) {
	    x = (*lidx) * R + R - 1;
	} else {
	    x = ((*lidx) * R) / (1<<((*lay)-1)) + R / (1<<(*lay)) - 1;
	}
    else {
	x = idx;
    }
    /*
     * This express the actual packet organization.
     */
#if 0		/* Mon Dec  8 12:26:13 GMT 1997 */
    *pkt = ( x%R + (x/(B*R))*R + x/(k*R*B) )%(k*B);
#else
    *pkt = ( x%R + (x/(B*R))*R + x/(k*R*B) )%(k*R);
#endif
    *blk = ( x + x/max(B, R))%B;

    if (idx < 0)
	return (x);

    /* XXX  compute lay, to be made more efficient */
    {
	int div = 2, sub = 0, i = 0;
	while ( (x-sub)%div != 0) {
	    if (i!=(nc-2)) div *= 2;
	    sub += (1<<i);
	    i++;
	}
	if (*lay >= 0) {
	    NDEB(fprintf(stderr, "*lay %d computed %d idx %d\n",
		*lay, nc-i-1, x) );
#if 0	/* doesn't work because we dont take into account bursts */
	    assert (*lay == nc-i-1);
#endif
	} else
	    *lay = nc-i-1;
	return (x);
    }
}

int 
compute_k_B(int *k, int *B, int npkts, int nlay)
{
  int R = 1<<(nlay-1);

    *k = THE_K;
    if (*k > K_MAX) return(-1);
    if (R > N_MAX) return(-1);
    *B = (npkts + *k - 1)/ *k;
    if (*B < R) {
	*B = R / (R / *B);
    } else {
	*B = ((*B - 1)/R + 1) * R;
    }
    return(0);
}

